# http://www.neo-tech-lab.co.uk/MikuMikuDance/PMD1.htm

def read_str(f, n)
  NKF.nkf('-Se', f.read(n).sub(/\0.*/, ''))
end

def read_int(f)
  f.read(4).unpack("V")[0]
end

def read_short(f)
  f.read(2).unpack("v")[0]
end

def read_byte(f)
  f.read(1).unpack("C")[0]
end

require 'nkf'

pmd_filename = ARGV[0] || 'miku.pmd'
f = File.open(pmd_filename)

magic = f.read(3)
puts "magic: #{magic}"

version = f.read(4).unpack("f")[0]
puts "version: #{version}"

name = NKF.nkf('-Se', f.read(20).sub(/\0.*/, ''))
puts "name: #{name}"

comment = NKF.nkf('-Se', f.read(256).sub(/\0.*/, ''))
puts "comment:\n#{comment}\n"

num_vertices = f.read(4).unpack("V")[0]
puts "num_vertices: #{num_vertices}"
vertices = []
bbox = [[0, 0, 0], [0, 0, 0]]
num_vertices.times do
  v = f.read(38).unpack("f3f3f2v2cc")
  #p v[8, 3]
  vertices << v
  3.times{|i|
    bbox[0][i] = [bbox[0][i], vertices[-1][i]].min
    bbox[1][i] = [bbox[1][i], vertices[-1][i]].max
  }
end
puts 'bounding box: %s' % bbox.inspect
#p vertices

num_faces = f.read(4).unpack("V")[0]
puts "num_faces: #{num_faces}"
faces = f.read(2 * num_faces).unpack("v#{num_faces}")

num_materials = f.read(4).unpack("V")[0]
puts "num_materials: #{num_materials}"
check_num_faces = 0
materials = []
num_materials.times do
  m = []
  # face color
  m << f.read(16).unpack("f4")
  # power
  m << f.read(4).unpack("f1")
  # specular color
  m << f.read(12).unpack("f3")
  # emissive color
  m << f.read(12).unpack("f3")

  # toon index, edge flag
  m << f.read(2).bytes.to_a
  # face vertex count
  m << vc = f.read(4).unpack("V")[0]
  check_num_faces += vc

  m << f.read(20)
  p m

  materials << m
end
puts "check_num_faces: #{check_num_faces}"

num_bones = f.read(2).unpack("v")[0]
puts "num_bones: #{num_bones}"
bones = []
euc_bone_names = []
num_bones.times do
  name = f.read(20).sub(/\0.*/, '')
  name = NKF.nkf('-Sw', name)

  bone = [name] + f.read(2 + 2 + 1 + 2 + 4 * 3).unpack("vvcvf3")
  bones << bone

  euc_name = NKF.nkf('-We', name)
  euc_bone_names << euc_name
  print euc_name + '... '
  p bone[1..-1]
end

bones.zip(euc_bone_names). each do |bone, name|
  n, parent, tail, type, ik, x, y, z = *bone
  parent = euc_bone_names[parent] || 'ROOT'
  tail = euc_bone_names[tail] || 'LEAF'
  ik = euc_bone_names[ik] || ''
  puts "#{name}: parent=#{parent} tail=#{tail} ik=#{ik} type=#{type} pos=#{x},#{y},#{z}"
end

num_iks = f.read(2).unpack("v")[0]
puts "num_iks: #{num_iks}"
iks = []
num_iks.times do
  ik = f.read(2 + 2 + 1 + 2 + 4).unpack("vvcvf")
  ik[2] = f.read(ik[2] * 2).unpack("v#{ik[2]}")
  iks << ik
  p ik
  puts "#{euc_bone_names[ik[0]]} #{euc_bone_names[ik[1]]} [#{ik[2].map{|b|euc_bone_names[b]}*','}]"
end

num_skins = f.read(2).unpack("v")[0]
puts "num_skins: #{num_skins}"
skins = []
num_skins.times do
  name = read_str(f, 20)
  cnt = f.read(4).unpack("V")[0]
  type = f.read(1)
  a = []
  cnt.times do
    a << f.read(16).unpack("Vf3")
  end
  puts name

  skins << [name, a]
end

num_skin_disps = f.read(1).unpack("C")[0]
puts "num_skin_disps: #{num_skin_disps}"
skin_disps = f.read(2 * num_skin_disps).unpack("v#{num_skin_disps}")

num_bone_disp_names = f.read(1).unpack("C")[0]
puts "num_bone_disp_names: #{num_bone_disp_names}"
bone_disp_names = []
num_bone_disp_names.times do
  name = NKF.nkf('-Se', f.read(50).sub(/\0.*/, ''))
  puts name
  bone_disp_names << name
end

num_bone_disps = f.read(4).unpack("V")[0]
puts "num_bone_disps: #{num_bone_disps}"
bone_disps = []
num_bone_disps.times do
  bone_disps << f.read(3).unpack("vC")
end

has_eng_name = f.read(1).unpack("C")[0]
eng_name = read_str(f, 20)
eng_comment = read_str(f, 256)
num_bones.times do
  read_str(f, 20)
end
(num_skins-1).times do
  read_str(f, 20)
end
num_bone_disp_names.times do
  read_str(f, 50)
end

# toon textures
10.times do
  read_str(f, 100)
end

num_rigidbodies = read_int(f)
puts "num_rigidbodies: #{num_rigidbodies}"
bodies = []
num_rigidbodies.times do
  name = read_str(f, 20)
  body = [name, *f.read(63).unpack("vCvCf3f3f3fffffC")]

  print name, " ", euc_bone_names[body[1]], ": "
  p body[2..-1]

  bodies << body
end

num_joints = read_int(f)
puts "num_joints: #{num_joints}"
joints = []
num_joints.times do
  name = read_str(f, 20)
  joint = [name, *f.read(104).unpack("VVf3f3f3f3f3f3f3f3")]

  body1 = bodies[joint[1]]
  body2 = bodies[joint[2]]
  print name, " ", euc_bone_names[body1[1]], "-", euc_bone_names[body2[1]]
  p joint[3..-1]
  joints << joint
end


10.times{
  a = ''
  s = f.read(16)
  break if !s
  l = s.chars.map{|c|
    if c[0] < 32 || c[0] > 126
      a << '.'
    else
      a << c
    end
    c[0].to_s(16).ljust(2,"0")
  } * ' '

  puts "#{l}  #{a}"
}


#ratio = 400.0 / (bbox[1][0] - bbox[0][0])
ratio = 600.0 / (bbox[1][1] - bbox[0][1])
puts "ratio: #{ratio}"

js_vertices = []
vertices.each do |v|
  v = v[0, 3] + v[6..-1] + v[3, 3]
  js_vertices << v
end

js_faces = []
mi = 0
mmi = materials[0][-2] / 3
(num_faces/3).times do |i|
  f = faces[i*3,3]
  if i >= mmi
    mi += 1
    mmi += materials[mi][-2] / 3
  end
  f << mi
  js_faces << f
end

File.open(pmd_filename + '.js', 'w') do |of|
  of.puts 'VERTICES = ['
  js_vertices.each do |v|
    #x = (v[0] - bbox[0][0]) * ratio + 20
    #y = (bbox[1][1] - v[1]) * ratio + 20
    #z = (v[2] - bbox[0][2]) * ratio + 20

    of.print '[' + v*',' + ']'
    of.puts js_vertices[-1] == v ? '' : ','
  end
  of.puts '];'

  of.puts 'FACES = ['
  js_faces.each do |f|
    of.print '[' + f*',' + ']'
    of.puts js_faces[-1] == f ? '' : ','
  end
  of.puts '];'

  of.puts 'MATERIALS_GL = ['
  materials.each do |m|
    #c = m[0][0,3].map{|v|(v*255).to_i}
    #c << m[0][3]
    js_m = m[0]

    if m[-1][0] != ?\0
      img = m[-1].sub(/\0.*/,'')
      png = img + '.png'
      system("convert #{img} #{png}")
      js_m << "'" + png + "'"
    end
    of.print '[' + js_m * ',' + ']'
    of.puts materials[-1] == m ? '' : ','
  end
  of.puts '];'

  of.puts 'MATERIALS = ['
  materials.each do |m|
    if m[0][3] == 1.0
      c = m[0][0,3].map{|v|(v*255).to_i}
      js_m = ['\'rgb('+c*','+')\'']
    else
      c = m[0][0,3].map{|v|(v*255).to_i}
      c << m[0][3]
      js_m = ['\'rgba('+c*','+')\'']
    end
    if m[-1][0] != ?\0
      img = m[-1].sub(/\0.*/,'')
      png = img + '.png'
      system("convert #{img} #{png}")
      js_m << "'" + png + "'"
    end
    of.print '[' + js_m * ',' + ']'
    of.puts materials[-1] == m ? '' : ','

  end
  of.puts '];'

  of.puts 'BONES = ['
  bones.each do |b|
    b[0] = "'" + b[0] + "'"
    of.print '[' + b * ',' + ']'
    of.puts bodies[-1] == b ? '' : ','
  end
  of.puts '];'

  of.puts 'BODIES = ['
  bodies.each do |b|
    b[0] = "'" + NKF.nkf('-Ew', b[0]) + "'"
    of.print '[' + b * ',' + ']'
    of.puts bodies[-1] == b ? '' : ','
  end
  of.puts '];'

  of.puts 'JOINTS = ['
  joints.each do |j|
    j[0] = "'" + NKF.nkf('-Ew', j[0]) + "'"
    of.print '[' + j * ',' + ']'
    of.puts joints[-1] == j ? '' : ','
  end
  of.puts '];'

  of.puts 'SKINS = ['
  skins.each do |s|
    s[0] = "'" + NKF.nkf('-Ew', s[0]) + "'"
    s[1] = s[1].inspect
    of.print '[' + s * ',' + ']'
    of.puts skins[-1] == s ? '' : ','
  end
  of.puts '];'
end

exit


ratio = 400.0 / (bbox[1][1] - bbox[0][1])

require 'sdl'

SDL.init(SDL::INIT_VIDEO)

scr = SDL.setVideoMode(640, 480, 32, SDL::SWSURFACE)

color = scr.format.mapRGB(255, 255, 255)

vertices.each do |v|
  x = (v[0] - bbox[0][0]) * ratio + 120
  #y = (v[1] - bbox[0][1]) * ratio + 40
  y = (bbox[1][1] - v[1]) * ratio + 40
  #puts "#{x} #{y}"
  scr.fillRect(x, y, 1, 1, color)
end

scr.flip
while true
  while ev = SDL::Event2.poll
    case ev
    when SDL::Event2::Quit
      exit
    when SDL::Event2::KeyDown
      if ev.sym == SDL::Key::ESCAPE
        exit
      end
    end
  end
  sleep 0.1
end
